// @flow
import React, { PureComponent } from 'react';
import { Image, Text, TouchableWithoutFeedback, View, ActivityIndicator, Platform } from 'react-native';
import { ColorSchemeContext, contextualStyleSheet, isDarkMode } from 'react-native-contextual-dark-mode';

import AnimatedCircularProgress from './AnimatedCircularProgress';
import WarningIcon from '../assets/warning.svg';

type ImageWithSpinnerProps = {
    imageUri: string;
    onPress: () => void;
    initialIsLoading?: boolean;
    errorTitle?: string;
    errorMessage?: string;
};

export class ImageWithSpinner extends PureComponent<ImageWithSpinnerProps> {
    static contextType = ColorSchemeContext;
    static defaultProps = {
        initialIsLoading: true,
        errorTitle: 'Could not load the picture',
        errorMessage: 'Make sure you have an internet connection and try again',
    };

    state = {
        isLoading: this.props.initialIsLoading,
        hasError: false,
    };
    errorMessage = null;
    animatedCircularProgressRef: ?AnimatedCircularProgress;

    handleDidLoad = () => this.setState({ isLoading: false, hasError: false });

    handleLoadError = ({ nativeEvent }) => {
        console.log(`Could not load image ${this.props.imageUri} with error: ${nativeEvent.error}`);
        this.errorMessage = nativeEvent.error;
        this.setState({ isLoading: false, hasError: true });
    }

    handleLoadProgressIOS = ({ nativeEvent }) => {
        const progress = nativeEvent.loaded / nativeEvent.total;
        this.animatedCircularProgressRef?.animate(progress * 95 + 2, 100);
    }

    setAnimatedCircularProgressRef = (ref) => {
        this.animatedCircularProgressRef = ref;
    }

    render() {
        return (
            <TouchableWithoutFeedback onPress={this.props.onPress}>
                <View style={s.flex1}>
                    <Image
                        source={{ uri: this.props.imageUri }}
                        style={s.flex1}
                        resizeMode="contain"
                        onError={this.handleLoadError}
                        onLoad={this.handleDidLoad}
                        onProgress={this.handleLoadProgressIOS}
                    />
                    <View style={s.imageLoaderContainer}>
                        {this.state.isLoading && (Platform.OS === 'ios' ? (
                            <AnimatedCircularProgress
                                ref={this.setAnimatedCircularProgressRef}
                                size={70}
                                width={6}
                                fill={3}
                                backgroundWidth={7}
                                backgroundColor={isDarkMode() ? '#FFFFFF50' : '#80808050'}
                                tintColor={isDarkMode() ? 'white' : '#808080'}
                                rotation={1}
                                lineCap="round"
                                spinBeforeFill
                            />
                        ) : (
                            <ActivityIndicator
                                style={s.spinner}
                                color={isDarkMode() ? 'white' : '#808080'}
                            />
                        ))}
                        {this.state.hasError && (
                            <>
                                <WarningIcon width={50} height={50} fill="#CE2C53" />
                                <Text style={s.imageLoaderErrorTitle}>
                                    {this.props.errorTitle}
                                </Text>
                                <Text style={s.imageLoaderErrorMessage}>
                                    {this.props.errorMessage}
                                </Text>
                            </>
                        )}
                    </View>
                </View>
            </TouchableWithoutFeedback>
        );
    }
}

const s = contextualStyleSheet((darkMode) => ({
    flex1: { flex: 1 },
    spinner: { position: 'absolute' },
    imageLoaderContainer: {
        position: 'absolute',
        height: '100%',
        width: '100%',
        alignItems: 'center',
        justifyContent: 'center',
        padding: 50,
    },
    imageLoaderErrorTitle: {
        color: darkMode ? 'white' : '#808080',
        fontSize: 17,
        fontWeight: 'bold',
        marginTop: 30,
        width: '100%',
        textAlign: 'center',
    },
    imageLoaderErrorMessage: {
        color: darkMode ? 'white' : '#808080',
        fontSize: 15,
        marginTop: 30,
        width: '100%',
        textAlign: 'center',
    },
}));
