// @flow
import React, {Component} from 'react';
import {
    ActivityIndicator,
    Animated,
    CompositeAnimation,
    Dimensions,
    Easing,
    FlatList,
    GestureResponderEvent,
    NativeTouchEvent,
    Platform,
    StyleProp,
    StyleSheet,
    View,
} from 'react-native';
import type {LayoutEvent} from 'react-native/Libraries/Types/CoreEventTypes';
import {delta, distance, isValidIndex, meanPoint, touchToFinger} from "./helpers/pinchZoomHelper";
import {log} from "./helpers/loggingHelper";

type Props<Item> = {
    style?: StyleProp,
    zoomDisabled?: boolean,
    pictureWidth?: number,
    pictureHeight?: number,
    onZoomStart?: () => void,
    onZoomEnd?: () => void,
    onDragDown?: () => void,
    minZoom?: number,
    maxZoom?: number,
    dragDownDistance?: number,
    items: any[],
    // eslint-disable-next-line no-undef
    renderItem: (item: Item, index: number, isFocused: boolean, disableScroll: () => void, enableScroll: () => void) => React$Node,
    index?: number, // PinchZoomCarousel is not a controlled component because the `index` prop is not required
    onIndexChange?: (index: number) => void,
    keyExtractor?: (item: Item, index: number) => any;
    windowSize?: number,
};
export type Point = { x: number, y: number };
export type Finger = { x: number, y: number, id: string };

const ZOOM_EPSILON = 0.18;
const MEDIA_CAROUSEL_MARGIN = 40; // pts
const INERTIAL_DRAG_DECELERATION = 0.008; // pts*ms^-2
const INERTIAL_DRAG_DECELERATION_OUTSIDE_BOUNDS = 0.03; // pts*ms^-2
const SPRING_BACK_ANIMATION = { toValue: 1, duration: 400 /* ms */, useNativeDriver: false, easing: Easing.bezier(0.32, 0.61, 0.69, 0.98) };
const SPRING_BACK_ANIMATION_FAST = { ...SPRING_BACK_ANIMATION, duration: 100 };
const IS_ANDROID = Platform.OS === 'android';
const USE_OPTIMIZED_ROTATION_HANDLING = true;

const styles = StyleSheet.create({
    containerDefaultStyle: { flex: 1, backgroundColor: 'black' },
    scrollView: { flex: 1 },
    wrapper: { flex: 1 },
    loaderContainer: { ...StyleSheet.absoluteFill, backgroundColor: "black", alignItems: "center", justifyContent: "center" },
});

class PinchZoomCarousel<Item> extends Component<Props<Item>> {
    static defaultProps = {
        style: styles.containerDefaultStyle,
        minZoom: 0.6,
        maxZoom: 8,
        dragDownDistance: 150,
        keyExtractor: (item, index) => `PinchZoomCarouselChild${index}`,
        windowSize: 5,
    };

    container: ?View;
    flatList: ?FlatList;
    childrenRef: View[] = [];
    // gesture vars
    isMultiTouch = false;
    isDragDownGesture = false;
    screenspaceInitialTwoFingerDistance = 1;
    screenspaceGestureStartPoint: ?Point;
    screenspaceBeforeResponderPoint: ?Point = null;
    screenspaceVelocityX = 0; // points/millisecond
    screenspaceVelocityY = 0; // points/millisecond
    screenspaceLastTouchX = 0;
    screenspaceLastTouchY = 0;
    scrollOffsetX = 0;
    isScrollingBetweenMedia = false;
    // transforms
    scale = 1;
    childspaceTranslateX = 0;
    childspaceTranslateY = 0;
    fadeFactor = 1; // 1 = no fade, 0 = fully faded
    // gesture reference
    childspaceGestureInitialTranslateX = 0;
    childspaceGestureInitialTranslateY = 0;
    // layout
    childWidth = 1;
    childHeight = 1;
    // animations
    springAnimation: ?CompositeAnimation;
    inertialAnimation: ?CompositeAnimation;
    runOnUpdate: ?() => void;

    constructor(props) {
        super(props);
        const { width, height } = Dimensions.get('screen');
        this.state = {
            containerWidth: width,
            isPortrait: true, // typical case
            index: props.index || 0,
            scrollEnabled: true,
            loaded: true,
            isZooming: false,
        };
        this.containerHeight = height;
    }

    // eslint-disable-next-line camelcase
    UNSAFE_componentWillReceiveProps(nextProps: Props) {
        // fix black screen when you delete the last item on Android
        if (IS_ANDROID && this.state.index >= nextProps.items.length && nextProps.items.length > 0) {
            this.state.index = isValidIndex(nextProps.index) ? nextProps.index : nextProps.items.length - 1;
            this.scrollToIndex(this.state.index, false);
        }
        // update state following props
        if (isValidIndex(nextProps.index) && nextProps.index !== this.state.index) {
            this.state.index = nextProps.index;
            this.scrollToIndex(this.state.index, false);
        }
        // protection for out-of-bounds index when the list length changes and the parent does not update `index` prop correctly
        if (nextProps.items.length - 1 < this.state.index) {
            this.state.index = Math.max(nextProps.items.length - 1, 0);
        }
    }

    scrollToComputedOffset = () => {
        const offset = (this.state.containerWidth + MEDIA_CAROUSEL_MARGIN) * this.state.index;
        if (this.flatList) {
            log(`scrollToComputedOffset | scrolling to ${this.state.index} * (${this.state.containerWidth} + ${MEDIA_CAROUSEL_MARGIN}) = ${offset}`);
            this.flatList.scrollToOffset({ offset, animated: false });
        } else {
            log('scrollToComputedOffset | no flatList ref');
        }
        this.setState({ loaded: true });
    }

    /**
     * Public methods
     */

    /*
     * Resets the zoom level to 100% and removes any translation
     */
    resetZoomAndDrag = (withAnimation?: boolean) => {
        if (withAnimation) {
            this.startChildSpringAnimation(0, 0, 1);
        } else {
            this.childspaceTranslateX = 0;
            this.childspaceTranslateY = 0;
            this.scale = 1;
            this.applyTransform();
        }
    };

    resetAllZooms = () => {
        this.scale = 1;
        this.setState({ isZooming: false });
        this.childrenRef.forEach((child) => {
            child?.setNativeProps?.({ style: { transform: [{ scale: 1 }] } });
        });
    }

    /**
     * Perform a zoom-in or zoom-out animation
     */
    doubleTap = () => {
        if (this.scale > 1) {
            // zoom out
            this.startChildSpringAnimation(0, 0, 1);
        } else {
            // zoom in
            this.startChildSpringAnimation(0, 0, this.scale + 1);
        }
    };

    animateToNextIndex = () => {
        this.setScrollEnabled(true);
        this.scrollToIndex(this.state.index + 1, true);
    }

    /**
     * Go to the next media
     */
    incrementIndex = () => {
        if (this.state.index < this.props.items.length - 1) {
            if (this.scale > 1) {
                // animate zoom out
                this.startChildSpringAnimation(0, 0, 1, this.animateToNextIndex, true);
            } else {
                this.scale = 1;
                this.childspaceTranslateX = 0;
                this.childspaceTranslateY = 0;
                this.applyTransform();
                this.animateToNextIndex();
            }
        }
    }

    animateToPreviousIndex = () => {
        this.setScrollEnabled(true);
        this.scrollToIndex(this.state.index - 1, true);
    }

    /**
     * Go the the previous media
     */
    decrementIndex = () => {
        if (this.state.index > 0) {
            if (this.scale > 1) {
                // animate zoom out
                this.startChildSpringAnimation(0, 0, 1, this.animateToPreviousIndex, true);
            } else {
                this.scale = 1;
                this.childspaceTranslateX = 0;
                this.childspaceTranslateY = 0;
                this.applyTransform();
                this.animateToPreviousIndex();
            }
        }
    }

    /**
     * Private methods
     */

    /**
     * Animates the child view to the given (targetX, targetY) and targetScale.
     * Also handles the fadeFactor animation.
     * @param targetX: x in childspace at the end of the animation
     * @param targetY: y in childspace at the end of the animation
     * @param targetScale: scale at the end of the animation
     * @param onEnd: callback when the animation is done
     * @param fast: true to run the animation much faster
     */
    startChildSpringAnimation = (targetX: number, targetY: number, targetScale: number, onEnd?: ({ finished: boolean }) => void, fast?: boolean) => {
        const springBackProgress = new Animated.Value(0);
        const initialX = this.childspaceTranslateX;
        const initialY = this.childspaceTranslateY;
        const initialFade = this.fadeFactor;
        // targetFade = 1
        const initialScale = this.scale;
        springBackProgress.addListener(({ value: progress }) => {
            const inverseProgress = 1 - progress;
            this.scale = (inverseProgress * initialScale) + (progress * targetScale);
            this.childspaceTranslateX = (inverseProgress * initialX) + (progress * targetX);
            this.childspaceTranslateY = (inverseProgress * initialY) + (progress * targetY);
            this.fadeFactor = (inverseProgress * initialFade) + progress;
            this.applyTransform();
        });
        this.springAnimation = Animated.timing(springBackProgress, fast ? SPRING_BACK_ANIMATION_FAST : SPRING_BACK_ANIMATION);
        this.springAnimation.start(onEnd);
    };

    /**
     * Call this function when the user lift their fingers and the child view needs to continue moving by inertia.
     * The view will continue with the same speed and slow down physically by INERTIAL_DRAG_DECELERATION.
     * If the child view is outside bounds (outside the normal frame), the deceleration increases to INERTIAL_DRAG_DECELERATION_OUTSIDE_BOUNDS.
     * @param onEnd: callback when the animation is done
     */
    startChildInertialAnimation = (onEnd?: ({ finished: boolean }) => void) => {
        const animationTime = new Animated.Value(0);
        const initialTranslateX = this.childspaceTranslateX;
        const initialTranslateY = this.childspaceTranslateY;
        const initialVelocityX = this.screenspaceVelocityX;
        const initialVelocityY = this.screenspaceVelocityY;
        // target velocity is (0, 0) and decreases linearly
        const velocityMagnitude = Math.sqrt(initialVelocityX * initialVelocityX + initialVelocityY * initialVelocityY);
        if (velocityMagnitude > 0) {
            // is current position outside bounds or on bounds edge?
            const childspaceBorderX = (this.scale - 1) * this.childWidth / 2 / this.scale;
            const childspaceBorderY = (this.scale - 1) * this.childHeight / 2 / this.scale;
            const isOutsideBoundsX = Math.abs(this.childspaceTranslateX) >= childspaceBorderX;
            const isOutsideBoundsY = Math.abs(this.childspaceTranslateY) >= childspaceBorderY;
            // use more deceleration if that's the case
            const deceleration = (isOutsideBoundsX || isOutsideBoundsY) ? INERTIAL_DRAG_DECELERATION_OUTSIDE_BOUNDS : INERTIAL_DRAG_DECELERATION;
            const duration = velocityMagnitude / deceleration;
            const accelerationX = -deceleration * (initialVelocityX / velocityMagnitude);
            const accelerationY = -deceleration * (initialVelocityY / velocityMagnitude);
            animationTime.addListener(({ value: t }) => {
                const screenspaceOffsetX = initialVelocityX * t + (accelerationX * t * t) / 2;
                const screenspaceOffsetY = initialVelocityY * t + (accelerationY * t * t) / 2;
                this.childspaceTranslateX = initialTranslateX + (screenspaceOffsetX / this.scale);
                this.childspaceTranslateY = initialTranslateY + (screenspaceOffsetY / this.scale);
                this.applyTransform();
            });
            this.inertialAnimation = Animated.timing(animationTime, { toValue: duration, duration, useNativeDriver: false });
            this.inertialAnimation.start(onEnd);
        } else {
            onEnd?.({ finished: true });
        }
    };

    /**
     * Compute childWidth and childHeight to have a better idea of the bounds.
     * If the pictureWidth and pictureHeight are not given in props, defaults to containerWidth and containerHeight.
     */
    updateChildDimensions = (width, height) => {
        if (this.props.pictureWidth > 0 && this.props.pictureHeight > 0) {
            const childAspect = this.props.pictureWidth / this.props.pictureHeight;
            const containerAspect = width / height;
            // assuming resizeMode="cover"
            if (childAspect > containerAspect) {
                // wider child than container -> width is kept, child height < container height
                this.childWidth = width;
                this.childHeight = height * (containerAspect / childAspect);
            } else {
                // taller child than container -> height is kept, child width < container width
                this.childHeight = height;
                this.childWidth = width * (childAspect / containerAspect);
            }
        } else {
            this.childWidth = width;
            this.childHeight = height;
        }
    };

    setScrollEnabled = (scrollEnabled: boolean) => {
        if (this.state.scrollEnabled !== scrollEnabled) {
            this.setState({...this.state, scrollEnabled})
        }
    }

    enableScroll = () => this.setScrollEnabled(true);

    disableScroll = () => this.setScrollEnabled(false);

    /**
     * Gesture methods
     */

    /**
     * Set the appropriate state to begin handling a single touch gesture.
     * We save the position of the touch on screen and the current child (x,y) offsets
     * @param event
     */
    startMonoTouchGesture = (event: GestureResponderEvent) => {
        const touch = event.nativeEvent.touches[0];
        this.isMultiTouch = false;
        this.screenspaceGestureStartPoint = touchToFinger(touch);
        this.screenspaceLastTouchX = this.screenspaceGestureStartPoint.x;
        this.screenspaceLastTouchY = this.screenspaceGestureStartPoint.y;
        this.lastTouchTime = touch.timestamp;
        this.childspaceGestureInitialTranslateX = this.childspaceTranslateX;
        this.childspaceGestureInitialTranslateY = this.childspaceTranslateY;
    }

    /**
     * Set the appropriate state to begin handling a multi touch gesture.
     * We save the position of the touches on screen, the finger distance and the current child (x,y) offsets
     * @param event
     */
    startMultiTouchGesture = (event: GestureResponderEvent) => {
        const { touches } = event.nativeEvent;
        this.isMultiTouch = true;
        if (this.scale === 1) {
            this.props.onZoomStart?.();
        }
        const [finger1, finger2] = touches.map(touchToFinger);
        this.screenspaceInitialTwoFingerDistance = distance(finger1, finger2) / this.scale;
        this.screenspaceGestureStartPoint = meanPoint(finger1, finger2);
        this.childspaceGestureInitialTranslateX = this.childspaceTranslateX;
        this.childspaceGestureInitialTranslateY = this.childspaceTranslateY;
    }

    /**
     * Move the ScrollView to display the media at the current index.
     * @param index
     * @param animated
     */
    scrollToIndex = (index: number, animated = false) => {
        this.flatList?.scrollToIndex({ index, animated });
    }

    handleLayout = (event: LayoutChangeEvent) => {
        const { width, height } = event.nativeEvent.layout;
        this.onDimensionsChanged(width, height);
        this.updateChildDimensions(width, height);
    }

    /**
     * Touch and layout handlers
     */

    onDimensionsChanged = (width, height) => {
        log(`onDimensionsChanged | width = ${width} height = ${height}`);

        if (this.state.containerWidth === width || this.containerHeight === height) {
            log('same dimensions, ignoring');
            return;
        }

        this.setState({ containerWidth: width, loaded: false });
        this.containerHeight = height;

        this.updateChildDimensions(width, height);

        this.isChangingOrientation = true;
        setTimeout(this.scrollToComputedOffset, IS_ANDROID ? 100 : 0);
        this.resetZoomAndDrag();
        this.resetAllZooms();
    }

    /**
     * Same implementation as handleMoveShouldSetResponder.
     * @param event
     * @returns {boolean} whether the child view should become the responder
     */
    handleStartShouldSetResponder = (event: GestureResponderEvent) => {
        return this.handleMoveShouldSetResponder(event);
    }

    /**
     * Called when the child view is touched and may become responder for the gesture
     * @param event
     * @returns {boolean} whether the child view should become the responder
     */
    handleMoveShouldSetResponder = (event: GestureResponderEvent) => {
        const { touches } = event.nativeEvent;
        if (touches.length === 1) {
            // For touches with one finger, the (x,y) screen coordinates are first recorded.
            // A. If zooming, become responder. The child view will animate.
            // B. If not zooming, dont become responder and let the parent scroll view take over.
            //    Unless the user dragged down for at least 30 pts, then become responder for a drag down gesture.
            if (!this.screenspaceBeforeResponderPoint) {
                this.screenspaceBeforeResponderPoint = touchToFinger(touches[0]);
                return (this.scale - ZOOM_EPSILON) > 1;
            }
            const dy = touches[0].pageY - this.screenspaceBeforeResponderPoint.y;
            if (dy >= 30) {
                this.isDragDownGesture = true;
                this.fadeFactor = 1;
                return true;
            }
            this.isDragDownGesture = false;
            return (this.scale - ZOOM_EPSILON) > 1;
        }
        this.screenspaceBeforeResponderPoint = null;
        return !this.props.zoomDisabled;
    }

    // Always give up the responder
    handleResponderTerminationRequest = () => true;

    /**
     * The component is now the responder for the gesture.
     * Start a single or multi-touch gesture
     * @param event
     */
    handleGestureStart = (event: GestureResponderEvent) => {
        // this.setScrollEnabled(this.isScrollingBetweenMedia);
        this.inertialAnimation?.stop();
        this.springAnimation?.stop();
        this.screenspaceVelocityX = 0;
        this.screenspaceVelocityY = 0;
        this.animationStartIndex = this.state.index;
        const { touches } = event.nativeEvent;
        if (touches.length === 1) {
            this.startMonoTouchGesture(event);
        } else {
            this.startMultiTouchGesture(event);
        }
    }

    /**
     * Update the view following the latest touch event in a multi-touch gesture
     * @param touches
     */
    handleDragAndZoom = (touches: NativeTouchEvent[]) => {
        const fingers = touches.map(touchToFinger);
        const [finger1, finger2] = fingers;
        // zoom (%)
        const currentDistance = distance(finger1, finger2);
        const zoom = currentDistance / this.screenspaceInitialTwoFingerDistance;
        this.scale = Math.max(this.props.minZoom, Math.min(this.props.maxZoom, zoom));
        // drag (x y)
        const currentMean = meanPoint(finger1, finger2);
        const gestureDelta = delta(this.screenspaceGestureStartPoint, currentMean);
        this.childspaceTranslateX = gestureDelta.x / this.scale + this.childspaceGestureInitialTranslateX;
        this.childspaceTranslateY = gestureDelta.y / this.scale + this.childspaceGestureInitialTranslateY;
        // manipulate
        this.applyTransform();
    }

    /**
     * Update the view following the latest touch event in a mono-touch gesture
     * @param event
     */
    handleDragOnly = (event: GestureResponderEvent) => {
        const touch = event.nativeEvent.touches[0];
        const finger = touchToFinger(touch);
        const screenspaceGestureDelta = delta(this.screenspaceGestureStartPoint, finger);
        // Move the current child view
        this.childspaceTranslateX = screenspaceGestureDelta.x / this.scale + this.childspaceGestureInitialTranslateX;
        this.childspaceTranslateY = screenspaceGestureDelta.y / this.scale + this.childspaceGestureInitialTranslateY;
        if (this.isDragDownGesture) {
            // update fade
            this.fadeFactor = Math.min(1 - (screenspaceGestureDelta.y / this.containerHeight), 1);
            if (screenspaceGestureDelta.y > this.props.dragDownDistance) {
                this.props.onDragDown?.();
            }
        }
        // update velocity
        const dt = touch.timestamp - this.lastTouchTime;
        const dx = finger.x - this.screenspaceLastTouchX;
        const dy = finger.y - this.screenspaceLastTouchY;
        this.screenspaceVelocityX = dx / dt;
        this.screenspaceVelocityY = dy / dt;
        // update fields for next call
        this.screenspaceLastTouchX = finger.x;
        this.screenspaceLastTouchY = finger.y;
        this.lastTouchTime = touch.timestamp;
        this.applyTransform();
    }

    /**
     * Update the view following the latest touch event.
     * Detect if the gesture needs to transition to a mono-touch or multi-touch gesture,
     * otherwise update the view according to the current gesture mode.
     * @param event
     */
    handleGestureMove = (event: GestureResponderEvent) => {
        if (!event.nativeEvent) return;
        const { touches } = event.nativeEvent;
        if (this.isMultiTouch) {
            if (touches.length === 1) {
                // transition to mono-touch gesture
                this.startMonoTouchGesture(event);
            } else {
                // handle multi-touch event
                this.handleDragAndZoom(touches);
            }
        } else if (touches.length >= 2 && !this.props.zoomDisabled) {
            // transition to multi-touch gesture
            this.startMultiTouchGesture(event);
        } else {
            // handle mono-touch event
            this.handleDragOnly(event);
        }
    };

    /**
     * Called when the child view has stopped moving after an inertial animation.
     * If the zoom level is below 1 or close to 1, consider the zoom finished and reset the view.
     * If the view was dragged outside bounds, animate it back inside bounds.
     */
    handleChildInertialScrollEnd = () => {
        this.isDragDownGesture = false;
        // spring animate if the image was dragged outside bounds, keep the zoom level if above 100%
        if (this.scale - ZOOM_EPSILON < 1) {
            this.startChildSpringAnimation(0, 0, 1, ({ finished }) => {
                if (finished && this.scale === 1) {
                    this.props.onZoomEnd?.();
                }
            });
        } else {
            // return;
            const childspaceBorderX = (this.scale - 1) * this.childWidth / 2 / this.scale;
            const childspaceBorderY = (this.scale - 1) * this.childHeight / 2 / this.scale;
            const isOutsideBoundsX = Math.abs(this.childspaceTranslateX) > childspaceBorderX;
            const isOutsideBoundsY = Math.abs(this.childspaceTranslateY) > childspaceBorderY;
            if (isOutsideBoundsX || isOutsideBoundsY) {
                const targetX = isOutsideBoundsX ? Math.sign(this.childspaceTranslateX) * childspaceBorderX : this.childspaceTranslateX;
                const targetY = isOutsideBoundsY ? Math.sign(this.childspaceTranslateY) * childspaceBorderY : this.childspaceTranslateY;
                this.springAnimation?.stop();
                this.startChildSpringAnimation(targetX, targetY, this.scale);
            }
        }
    };

    /**
     * Called when the gesture has finished (i.e. the user has lifted all fingers or we lost the responder).
     * Start the inertial animation and reset the responder state to allow the ScrollView to handle the next touch (if applicable)
     */
    handleGestureEnd = () => {
        this.screenspaceBeforeResponderPoint = null;
        this.startChildInertialAnimation(this.handleChildInertialScrollEnd);
    };

    /**
     * Called when the ScrollView is responder and moves.
     * It updates the current index and notifies the parent component when it changes.
     * @param event
     */
    handleScroll = (event) => {
        if (this.isChangingOrientation) {
            this.isChangingOrientation = false;
            return;
        }

        this.scrollOffsetX = event.nativeEvent.contentOffset.x;
        const rawIndex = this.scrollOffsetX / (this.state.containerWidth + MEDIA_CAROUSEL_MARGIN);
        this.isScrollingBetweenMedia = this.scrollOffsetX % (this.state.containerWidth + MEDIA_CAROUSEL_MARGIN) !== 0;
        const index = Math.min(Math.max(Math.round(rawIndex), 0), this.props.items.length - 1);
        log(`handleScroll | contentOffsetX ${this.scrollOffsetX} index ${this.state.index} -> ${index}`);
        if (index !== this.state.index) {
            this.state.index = index;
            this.props.onIndexChange?.(index);
            this.resetAllZooms();
        }
    };

    /**
     * Reset the screenspaceBeforeResponderPoint to enable the drag down gesture detection.
     */
    resetBeforeResponderStartPoint = () => {
        this.screenspaceBeforeResponderPoint = null;
    }

    /**
     * Scales and translates the child view according to the current state.
     * Also updates the opacity of the container for the drag down animation.
     * This method can be called on every frame (or faster) because it does not cause a render.
     */
    applyTransform = () => {
        this.childrenRef[this.state.index]?.setNativeProps({
            transform: [
                { scale: this.scale * this.fadeFactor },
                { translateX: this.childspaceTranslateX },
                { translateY: this.childspaceTranslateY },
            ],
        });
        this.container?.setNativeProps({
            style: { opacity: this.fadeFactor },
        });
        this.setState({ isZooming: this.scale - ZOOM_EPSILON >= 1 });
    };

    setContainerRef = (ref) => { this.container = ref; };

    setFlatListRef = (ref) => { this.flatList = ref; };

    getItemLayout = (item: Item, index: number) => {
        const isLast = index === this.props.items.length - 1;
        const length = isLast ? this.state.containerWidth : this.state.containerWidth + MEDIA_CAROUSEL_MARGIN;
        const offset = (this.state.containerWidth + MEDIA_CAROUSEL_MARGIN) * index;
        return { length, offset, index };
    };

    renderItem = ({ item, index }) => {
        const marginRight = index === this.props.items.length - 1 ? 0 : MEDIA_CAROUSEL_MARGIN;
        const isFocused = this.state.index === index;
        return (
            <View
                ref={(ref) => { this.childrenRef[index] = ref; }}
                style={{ height: '100%', width: this.state.containerWidth, marginRight, overflow: 'hidden' }}
            >
                {this.props.renderItem(item, index, isFocused, this.disableScroll, this.enableScroll)}
            </View>
        );
    };

    render() {
        log(`render | index = ${this.state.index} snapToInterval = ${this.state.containerWidth} + ${MEDIA_CAROUSEL_MARGIN}`);
        return (
            <View
                key="PinchZoomCarouselContainer"
                ref={this.setContainerRef}
                style={this.props.style}
            >
                <View
                    style={styles.wrapper}
                    onStartShouldSetResponder={this.state.scrollEnabled && this.handleStartShouldSetResponder}
                    onMoveShouldSetResponder={this.state.scrollEnabled && this.handleMoveShouldSetResponder}
                    onResponderGrant={this.state.scrollEnabled && this.handleGestureStart}
                    onResponderTerminationRequest={this.state.scrollEnabled && this.handleResponderTerminationRequest}
                    onResponderTerminate={this.state.scrollEnabled && this.handleGestureEnd}
                    onResponderRelease={this.state.scrollEnabled && this.handleGestureEnd}
                    onResponderMove={this.state.scrollEnabled && this.handleGestureMove}
                    onLayout={this.handleLayout}
                >
                    <FlatList
                        // De-optimized mode: resets the FlatList when orientation changes to makes sure that `initialScrollIndex` is respected
                        // when the `snapToInterval` changes after a device rotation.
                        // Optimized mode will programmatically scroll to the correct index after a device rotation,
                        // but may not work on some devices.
                        key={USE_OPTIMIZED_ROTATION_HANDLING ? null : `PinchZoomCarouselFlatList-${this.state.isPortrait}`}
                        ref={this.setFlatListRef}
                        horizontal
                        initialScrollIndex={this.state.index}
                        style={styles.scrollView}
                        snapToInterval={this.state.containerWidth + MEDIA_CAROUSEL_MARGIN}
                        decelerationRate="fast"
                        disableIntervalMomentum // disallows swiping with momentum further that the next or previous item
                        removeClippedSubviews={false} // avoid <Image> load-in flashing (Android)
                        showsHorizontalScrollIndicator={false}
                        onScroll={this.handleScroll}
                        scrollEventThrottle={64} // 15 fps
                        onMomentumScrollBegin={this.resetBeforeResponderStartPoint}
                        onMomentumScrollEnd={this.resetBeforeResponderStartPoint}
                        data={this.props.items}
                        keyExtractor={this.props.keyExtractor}
                        getItemLayout={this.getItemLayout}
                        renderItem={this.renderItem}
                        windowSize={this.props.windowSize}
                        disableScrollViewPanResponder={false} // make sure that child touchable(s) are not triggered when you swipe
                        scrollEnabled={this.state.scrollEnabled && !this.state.isZooming}
                    />
                </View>
                {(IS_ANDROID && !this.state.loaded) &&
                    (<View style={styles.loaderContainer}>
                        <ActivityIndicator size="large" color="white" />
                    </View>)
                }
            </View>
        );
    }
}

export { PinchZoomCarousel };
